/* $Id: iterator.c 3 2007-07-09 07:59:45Z romain $ */

#include <regex.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "iterator.h"

regex_t iterator_preg ;
regmatch_t *iterator_match ;
int iterator_match_parts ;

int
iterator_initialize(void)
{
  int r = 0;
  jmp_buf jmp_env;

  switch (setjmp(jmp_env)) {
    case 0:
      if (regcomp(&iterator_preg, "^(.+)=(-?[[:digit:]]+)(\\.\\.-?[[:digit:]]+)?((\\+|-)[[:digit:]]+)?(:..+)?$", REG_EXTENDED)) {
        longjmp(jmp_env, 1);
      }
      if (!(iterator_match = malloc(sizeof(*iterator_match) * (1 + iterator_preg.re_nsub)))) {
        longjmp(jmp_env, 2);
      }
      r = iterator_match_parts = iterator_preg.re_nsub;
      break;
    case 2:
      regfree(&iterator_preg);
    case 1:
      r = 0;
      break;
  }

  return r;
}

void
iterator_finalize(void)
{
	regfree(&iterator_preg);
	if (iterator_match) {
		free(iterator_match);
		iterator_match = NULL;
	}
}

Iterator
iterator_new(char *iteration_domain)
{
  Iterator I;

  if (0 != regexec(&iterator_preg, iteration_domain, iterator_preg.re_nsub + 1 ,iterator_match, 0)) {
    fprintf(stderr,"Malformed iteration domain: %s\n", iteration_domain);
    return NULL;
  }

  if ((I = malloc(sizeof(*I)))) {
    I->name = NULL;

    I->start = 1;
    I->value = 1;
    I->has_stop = 0;
    I->stop = 0;
    I->stride = 1;
    I->format = NULL;

    I->next = NULL;
    I->prev = NULL;

    {
      int k;

      for (k=0; k <= iterator_preg.re_nsub; k++) {
        if (iterator_match[k].rm_so != -1) {
          size_t len = iterator_match[k].rm_eo - iterator_match[k].rm_so;
          char *s = malloc(len + 1);

          strncpy(s, iteration_domain + iterator_match[k].rm_so, len);
          s[len] = '\0';

#ifdef _LAZY_DEVELOPER
          printf("match[%d] = (%s)\n", k, s);
#endif
          switch (k) {
            case 1:
              I->name = strdup(s);
              break;
            case 2:
              sscanf(s, "%d", &(I->start));
              I->value = I->start;
              break;
            case 3:
              sscanf(s, "..%d", &(I->stop));
              I->has_stop = 1;
              break;
            case 4:
              sscanf(s, "%d", &(I->stride));
              break;
            case 6:
              I->format = strdup(s+1);
              break;
          }

          free(s);
        }
      }
    }

    /* Verify coherence */
    if (((I->stop - I->start) / I->stride) < 0) {
      fprintf(stderr, "Invalid iteration domain: %s. Check bounds and stride sign.\n", iteration_domain);
      iterator_delete(I);
      I = NULL;
    }
  }

  return I;
}

void
iterator_insert_after(Iterator I, Iterator new) {
  I->next = new;
  new->prev = I;
}

int
iterator_next(Iterator I)
{
  I->value += I->stride;
  return((double)(I->stop - I->value) / I->stride >= 0);
}

int
iterator_reset(Iterator I)
{
  return(I->value = I->start);
}

Iterator
iterator_delete(Iterator I)
{
  Iterator r = I->next;

  if (I->next) {
    I->next->prev = I->prev;
  }
  if (I->prev) {
    I->prev->next = I->next;
  }
  free(I->name);
  if (I->format) free(I->format);
  free(I);

  return r;
}


void
iterator_iterate(Iterator I, char *cmd, int *error)
{
  do {
    char buf[BUFSIZ];

    if (I->format) {
      sprintf(buf, I->format, I->value);
    } else {
      sprintf(buf, "%d", I->value);
    }
    setenv(I->name, buf, 1);

    if (I->next) {
      iterator_iterate(I->next, cmd, error);
    } else {
      *error = system(cmd);
    }
  } while ((!(*error)) && iterator_next(I));
  iterator_reset(I);
}
